                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*              context              */
                     /*-----------------------------------*/

#include "SlipMain.h"

#include "SlipContext.h"
#include "SlipEnvironment.h"
#include "SlipThread.h"

/*----------------------------------------------------------------------------*/
/*------------------------------ Memory management----------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------- private macros -----------------------------*/

#define BEGIN_ATOMIC                                                           \
  Concurrency_Mutex_Lock(&Concurrency_mutex)
  
#define END_ATOMIC                                                             \
  Concurrency_Mutex_Unlock(&Concurrency_mutex)

#define CONTINUE                                                               \
  Concurrency_Broadcast_Conditional(&Concurrency_conditional)

#define WAIT                                                                   \
  Concurrency_Wait_Conditional(&Concurrency_conditional,                       \
                               &Concurrency_mutex)
                                 
/*------------------------------ private constants ---------------------------*/

enum { Activate_id            ,
       Deactivate_id          ,
       Garbage_collect_id     ,
       Extend_global_frame_id ,
       Extend_context_table_id };

/*-------------------------------- private types -----------------------------*/

typedef 
  enum { Do_nothing              = 0                         ,
         Do_activate             = 1<<Activate_id            ,
         Do_deactivate           = 1<<Deactivate_id          ,
         Do_garbage_collect      = 1<<Garbage_collect_id     ,
         Do_extend_global_frame  = 1<<Extend_global_frame_id ,
         Do_extend_context_table = 1<<Extend_context_table_id } Sta_type;

/*------------------------------ private prototypes --------------------------*/

static NIL_type        atomic_garbage_collect(NIL_type);
static NIL_type    atomic_extend_global_frame(NIL_type);
static NIL_type atomic_extend_context_table_M(NIL_type);
    
/*------------------------------ private variables ---------------------------*/

static UNS_type Accumulated_claim;
static UNS_type Active_count;
static SPL_type Claim_spin;
static CND_type Concurrency_conditional;
static MTX_type Concurrency_mutex;
static SPL_type Freelist_spin;
static Sta_type Global_status;
static SPL_type Memory_spin;

/*------------------------------ private functions ---------------------------*/

static BYT_type memory_overflow(NIL_type)
  { return (Memory_Available() < Accumulated_claim); }

static NIL_type check_status(Sta_type Status)
  { if (memory_overflow())
      Global_status |= Do_garbage_collect;
    if ((Global_status == Do_nothing) && (Status == Do_nothing))
      return;
    BEGIN_ATOMIC;
      if (Status != Do_activate)
        { Active_count--;
          if (Status != Do_deactivate)
            Global_status |= Status; }
      if (Global_status != Do_nothing)      
        if (Active_count > 0)
          WAIT;
        else
          { if (Global_status & Do_garbage_collect)
              atomic_garbage_collect();
            if (Global_status & Do_extend_global_frame)
              atomic_extend_global_frame();
            if (Global_status & Do_extend_context_table)
              atomic_extend_context_table_M();
            Global_status = Do_nothing;
            CONTINUE; }
      if (Status != Do_deactivate)
        Active_count++; 
    END_ATOMIC; }
                       
/*----------------------------------------------------------------------------*/

static NIL_type atomic_garbage_collect(NIL_type)
  { Slip_Log("garbage collect");
    Memory_Collect();
    if (memory_overflow())
      Main_Fail(IMM_error); }
    
/*------------------------------ public functions ----------------------------*/

NIL_type Context_Collect_C(NIL_type)
  { check_status(Do_garbage_collect); }

PTR_type Context_Make_Chunk(BYT_type Tag,
                            UNS_type Size)
  { PTR_type pointer;
    Concurrency_Spin_Lock(&Memory_spin);
    pointer = Memory_Make_Chunk(Tag,
                                Size);
    Concurrency_Spin_Unlock(&Memory_spin);
    return pointer; }

NIL_type Context_Make_Claim(UNS_type Size)
  { Concurrency_Add(&Accumulated_claim,
                    Size);
    check_status(Do_nothing); }

NIL_type Context_Release_Claim(UNS_type Size)
  { Concurrency_Subtract(&Accumulated_claim,
                         Size); }

/*----------------------------------------------------------------------------*/

NIL_type Context_Memory_Initialize(NIL_type)
  { Concurrency_Initialize_Conditional(&Concurrency_conditional);
    Concurrency_Initialize_Mutex(&Concurrency_mutex);
    Concurrency_Initialize_Spin(&Claim_spin);
    Concurrency_Initialize_Spin(&Freelist_spin);
    Concurrency_Initialize_Spin(&Memory_spin);
    Accumulated_claim = 0;
    Global_status = Do_nothing; }

/*----------------------------------------------------------------------------*/
/*------------------------------ Context management---------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------ private constants ---------------------------*/

enum { Context_table_size_increment =  8,
         Initial_Context_table_size = 16 };

/*-------------------------------- private types -----------------------------*/

Define_Grammar_Frame(Ctx);

Begin_Grammar_Frame(Ctx);
  Define_Grammar_Slot(EXP, exp);
  Define_Grammar_Slot(ENV, env);
  Define_Grammar_Slot(FRM, frm);
  Define_Grammar_Slot(THR, thr);
  Define_Grammar_Slot(EXP, prt);
  Define_Grammar_Slot(TSK, tsk);
  Define_Grammar_Slot(EPL, epl);
  Define_Grammar_Slot(TPL, tpl);
End_Grammar_Frame(Ctx);
               
/*------------------------------ private variables ---------------------------*/

static VEC_type Context_table;
static CID_type Context_table_freelist;
static UNS_type Total_context_size;

/*------------------------------ private functions ---------------------------*/

static Ctx_type make_context_M(EXP_type Expression,
                               ENV_type Environment,
                               FRM_type Frame)
  { Ctx_type context;
    EPL_type environment_pool;
    PTH_type pthread;
    TPL_type thread_pool;
    TSK_type task;
    environment_pool = Environment_Setup_M();
    thread_pool = Thread_Setup_M();
    pthread = Concurrency_Self();
    task = make_TSK_M(pthread);
    context = (Ctx_type)make_VEC_M(Ctx_size);
    context->exp = Expression;
    context->env = Environment;
    context->frm = Frame;
    context->thr = Thread_Null;
    context->prt = Grammar_Null;
    context->tsk = task;
    context->epl = environment_pool;
    context->tpl = thread_pool;
    return context; }

static UNS_type context_size(NIL_type)
  { UNS_type size;
    size = size_TSK();
    size += Ctx_size + 2;
    return size; }
  
static Ctx_type get_context(CID_type Context_id)
  { Ctx_type context;
    context = Context_table[Context_id];
    return context; }
    
static NIL_type set_context(CID_type Context_id,
                            Ctx_type Context)
  { Context_table[Context_id] = Context; }
        
/*----------------------------------------------------------------------------*/

static NIL_type build_freelist(UNS_type Start)
  { CID_type context_id;
    NBR_type next_free;
    UNS_type size;
    size = size_VEC(Context_table);
    Context_table_freelist = Start + 1;
    for (context_id = Context_table_freelist;
         context_id < size;
         context_id++)
      { next_free = make_NBR(context_id + 1);
        Context_table[context_id] = next_free; }
    Context_table[size] = Grammar_Zero; }
    
static CID_type pop_freelist_C(NIL_type)
  { CID_type context_id,
             next_context_id;
    NBR_type next_free;
    UNS_type size;
    if (Context_table_freelist == 0)
      { size = size_VEC(Context_table);
        size += Context_table_size_increment + 1;
        Concurrency_Add(&Accumulated_claim,
                        size);
        check_status(Do_extend_context_table);
        Concurrency_Subtract(&Accumulated_claim,
                             size); }
    Concurrency_Spin_Lock(&Freelist_spin);
    context_id = Context_table_freelist;
    next_free = Context_table[context_id];
    next_context_id = get_NBR(next_free);
    Context_table_freelist = next_context_id; 
    Concurrency_Spin_Unlock(&Freelist_spin);
    return context_id; }
    
static NIL_type push_freelist(CID_type Context_id)
  { NBR_type next_free;
    Concurrency_Spin_Lock(&Freelist_spin);
    next_free = make_NBR(Context_table_freelist);
    Context_table[Context_id] = next_free;
    Context_table_freelist = Context_id; 
    Concurrency_Spin_Unlock(&Freelist_spin); }

/*----------------------------------------------------------------------------*/

static NIL_type atomic_extend_context_table_M(NIL_type)
  { UNS_type last;
    Slip_Log("extend context table");
    last = size_VEC(Context_table);
    Main_Reallocate_Vector_M(&Context_table,
                             Context_table_size_increment);
    build_freelist(last); }
    
/*------------------------------ public functions ----------------------------*/

EXP_type Context_Get_Expression(CID_type Context_id)
  { Ctx_type context;
    EXP_type value;
    context = get_context(Context_id);
    value = context->exp;
    return value; }

NIL_type Context_Flush(NIL_type)
  { Ctx_type context;
    EPL_type environment_pool;
    FRM_type frame;
    PTH_type pthread;
    TSK_type task;
    UNS_type context_id,
             size;
    context = get_context(Context_Main_Id);
    frame            = context->frm;
    environment_pool = context->epl;
    Environment_Release_Frame(frame,
                              environment_pool);
    Context_Reset_Environment(Context_Main_Id);
    size = size_VEC(Context_table);
    for (context_id = Context_Main_Id + 1;
         context_id <= size;
         context_id++)
      { context = get_context(context_id);
        if (!is_NBR(context))
          { context = get_context(context_id);
            task = context->tsk;
            pthread = task->pth;
            Concurrency_Cancel(pthread); }}
    build_freelist(Context_Main_Id); }

EXP_type Context_Restore(CID_type Context_id)
  { Ctx_type context;
    EXP_type expression;
    context = get_context(Context_id);
    expression = context->prt;
    context->prt = Grammar_Null;
    return expression; } 

NIL_type Context_Save(CID_type Context_id,
                      EXP_type Expression)
  { Ctx_type context;
    context = get_context(Context_id);
    context->prt = Expression; } 
                               
NIL_type Context_Set_Expression(CID_type Context_id,
                                EXP_type Expression)
  { Ctx_type context;
    context = get_context(Context_id);
    context->exp = Expression; }         

/*----------------------------------------------------------------------------*/

static NIL_type context_initialize_M(NIL_type)
  { Ctx_type context;
    Context_table = make_VEC_M(Initial_Context_table_size); 
    Main_Register(Context_table); 
    context = make_context_M(Grammar_Unspecified,
                             Environment_Global_Environment,
                             Environment_Global_Frame);
    set_context(Context_Main_Id,
                context);
    build_freelist(Context_Main_Id);
    Total_context_size = context_size() +
                         Environment_Pool_Size() +
                         Thread_Pool_Size();
    Active_count = 1; }

/*----------------------------------------------------------------------------*/
/*----------------------------- Fork/Join management--------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------ public functions ----------------------------*/

TXT_type Context_Read(TXT_type Text)
  { TXT_type input;
    check_status(Do_deactivate);
    input = Main_Read(Text);     /* resolve multiple reads in various pthread */
    check_status(Do_activate);
    return input; }

CID_type Context_Spawn_Task_C(CID_type Host_context_id,
                              SPR_type Start_procedure_C)
  { CID_type new_context_id;
    Ctx_type host_context,
             new_context;
    ENV_type environment;
    EXP_type expression;
    FRM_type frame;
    new_context_id = pop_freelist_C();
    Main_Claim_Dynamic(Total_context_size);
    host_context = get_context(Host_context_id);
    expression  = host_context->exp;
    environment = host_context->env;
    frame       = host_context->frm;
    apply_MRK(frame);
    new_context = make_context_M(expression,
                                 environment,
                                 frame);
    Main_Release_Dynamic(Total_context_size);
    set_context(new_context_id,
                new_context);
    Concurrency_Pthread(&(new_context->tsk->pth),
                        (WRK_type)Start_procedure_C,
                        (ADR_type)new_context_id);
    return new_context_id; }

NIL_type Context_Start_Task(NIL_type)
  { check_status(Do_activate); }
                                 
NIL_type Context_Stop_Task(NIL_type)
  { check_status(Do_deactivate);
    Concurrency_Kill(); }
                                 
EXP_type Context_Sync_Task(CID_type Context_id)
  { Ctx_type context;
    EXP_type value;
    PTH_type pthread;
    TSK_type task;
    context = get_context(Context_id);
    task = context->tsk;
    pthread = task->pth;
    check_status(Do_deactivate);
    Concurrency_Join(pthread);    /* resolve multiple joins with same pthread */
    check_status(Do_activate);
    context = get_context(Context_id);
    push_freelist(Context_id);
    value = context->exp;
    return value; }

/*----------------------------------------------------------------------------*/
/*---------------------------- Environment management ------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------ private functions ---------------------------*/

static NIL_type atomic_extend_global_frame(NIL_type)
  { Slip_Log("extend global frame");
    Environment_Confirm_Global_Frame(); }
    
/*------------------------------ public functions ----------------------------*/

NIL_type Context_Adjust_Global_Frame_C(UNS_type Size)
  { if (Environment_Adjust_Global_Frame_C(Size))
      check_status(Do_extend_global_frame); }
 
ENV_type Context_Environment_Grow_C(CID_type Context_id)
  { Ctx_type context;
    EPL_type environment_pool;
    ENV_type environment,
             new_environment;
    FRM_type frame;
    UNS_type size;
    context = get_context(Context_id);
    environment      = context->env;
    size = Environment_Get_Environment_size(environment);
    size += 1;
    Main_Claim_Dynamic(size);
    context = get_context(Context_id);
    environment      = context->env;
    environment_pool = context->epl;    
    new_environment = Environment_Make_M(size,
                                         environment_pool);
    Main_Release_Dynamic(size);
    frame = context->frm;
    Environment_Grow_Environment(new_environment,
                                 environment,
                                 frame);
    return new_environment; }

NIL_type Context_Environment_Replace(CID_type Context_id,
                                     ENV_type Environment,
                                     FRM_type Frame)
  { Ctx_type context;
    context = get_context(Context_id);
    context->env = Environment;
    context->frm = Frame; }

ENV_type Context_Get_Environment(CID_type Context_id)
  { Ctx_type context;
    ENV_type environment;
    context = get_context(Context_id);
    environment = context->env;
    return environment; }

FRM_type Context_Get_Frame(CID_type Context_id)
  { Ctx_type context;
    FRM_type frame;
    context = get_context(Context_id);
    frame = context->frm;
    return frame; }

NIL_type Context_Global_Frame_Set(EXP_type Value,
                                  UNS_type Offset)
  { Environment_Global_Frame[Offset] = Value; }

NIL_type Context_Global_Get(CID_type Context_id,
                            NBR_type Scope,
                            NBR_type Offset)                      
  { Ctx_type context;
    EXP_type value;
    ENV_type environment;
    UNS_type raw_scope,
             raw_offset;
    context = get_context(Context_id);
    environment = context->env;
    raw_scope = get_NBR(Scope);
    raw_offset = get_NBR(Offset);
    value = Environment_Global_Get(environment,
                                   raw_scope,
                                   raw_offset);
    context->exp = value; }
    
NIL_type Context_Global_Set(CID_type Context_id,
                            NBR_type Scope,
                            NBR_type Offset)
  { Ctx_type context;
    EXP_type value;
    ENV_type environment;
    UNS_type raw_offset,
             raw_scope;
    context = get_context(Context_id);
    value = context->exp;
    environment = context->env;
    raw_scope = get_NBR(Scope);
    raw_offset = get_NBR(Offset);
    Environment_Global_Set(environment,
                           value,
                           raw_scope,
                           raw_offset); }

NIL_type Context_Local_Get(CID_type Context_id,
                           NBR_type Offset)
  { Ctx_type context;
    EXP_type value;
    FRM_type frame;
    UNS_type raw_offset;
    context = get_context(Context_id);
    frame = context->frm;
    raw_offset = get_NBR(Offset);
    value = Environment_Local_Get(frame,
                                  raw_offset);
    context->exp = value; }

NIL_type Context_Local_Set(CID_type Context_id,
                           NBR_type Offset)
  { Ctx_type context;
    EXP_type value;
    FRM_type frame;
    UNS_type raw_offset;
    context = get_context(Context_id);
    value = context->exp;
    frame = context->frm;
    raw_offset = get_NBR(Offset);
    Environment_Local_Set(value,
                          frame,
                          raw_offset); }
                          
FRM_type Context_Make_Frame_C(CID_type Context_id,
                              UNS_type Size)
  { Ctx_type context;
    EPL_type environment_pool;
    FRM_type frame;
    Main_Claim_Dynamic(Size);
    context = get_context(Context_id);
    environment_pool = context->epl;
    frame = Environment_Make_Frame_M(Size,
                                     environment_pool);
    Main_Release_Dynamic(Size);
    return frame; }                               
                          
FRM_type Context_Mark_Frame(CID_type Context_id)
  { Ctx_type context;
    FRM_type frame;
    context = get_context(Context_id);
    frame = context->frm;
    apply_MRK(frame);
    return frame; }
    
NIL_type Context_Release_Current_Frame(CID_type Context_id)
  { Ctx_type context;
    EPL_type environment_pool;
    FRM_type frame;
    context = get_context(Context_id);
    frame            = context->frm;
    environment_pool = context->epl;
    Environment_Release_Frame(frame,
                              environment_pool); }

NIL_type Context_Release_Frame(CID_type Context_id,
                               FRM_type Frame)
  { Ctx_type context;
    EPL_type environment_pool;
    context = get_context(Context_id);
    environment_pool = context->epl;
    Environment_Release_Frame(Frame,
                              environment_pool); }

NIL_type Context_Reset_Environment(CID_type Context_id)
  { Ctx_type context;
    context = get_context(Context_id);
    context->env = Environment_Global_Environment;
    context->frm = Environment_Global_Frame; }
                                    
/*----------------------------------------------------------------------------*/
/*------------------------------- Thread management---------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------ public functions ----------------------------*/

THR_type Context_Mark_Thread(CID_type Context_id)
  { Ctx_type context;
    THR_type thread;
    context = get_context(Context_id);
    thread = context->thr;
    Thread_Mark(thread);
    return thread; }
    
THR_type Context_Thread_Keep_C(CID_type Context_id)
  { Ctx_type context;
    TPL_type thread_pool;
    THR_type thread;
    Main_Claim_Default();
    context = get_context(Context_id);
    thread      = context->thr;
    thread_pool = context->tpl;
    thread = Thread_Keep_M(thread,
                           thread_pool);
    Main_Release_Default();
    context->thr = thread;
    return thread; }

THR_type Context_Thread_Patch_C(CID_type Context_id,
                                NBR_type Thread_id)
  { Ctx_type context;
    TPL_type thread_pool;
    THR_type thread;
    Main_Claim_Default();
    context = get_context(Context_id);
    thread      = context->thr;
    thread_pool = context->tpl;
    thread = Thread_Patch_M(thread,
                            Thread_id,
                            thread_pool);
    Main_Release_Default();
    return thread; }

THR_type Context_Thread_Peek(CID_type Context_id)
  { Ctx_type context;
    THR_type thread;
    context = get_context(Context_id);
    thread = context->thr;
    return thread; }

THR_type Context_Thread_Poke_C(CID_type Context_id,
                               NBR_type Thread_id,
                               EXP_type Tail_position_status,
                               UNS_type Size)
  { Ctx_type context;
    THR_type old_thread,
             thread;
    TPL_type thread_pool;
    Main_Claim_Default();
    context = get_context(Context_id);
    thread_pool = context->tpl;
    thread = Thread_Make_M(Thread_id,
                           Tail_position_status,
                           Size,
                           thread_pool);
    Main_Release_Default();
    context = get_context(Context_id);
    old_thread = context->thr;
    Thread_Poke(thread,
                old_thread,
                thread_pool);
    context->thr = thread;
    return thread; }
        
THR_type Context_Thread_Push_C(CID_type Context_id,
                               NBR_type Thread_id,
                               EXP_type Tail_position_status,
                               UNS_type Size)
  { Ctx_type context;
    THR_type old_thread,
             thread;
    TPL_type thread_pool;
    Main_Claim_Default();
    context = get_context(Context_id);
    thread_pool = context->tpl;
    thread = Thread_Make_M(Thread_id,
                           Tail_position_status,
                           Size,
                           thread_pool);
    Main_Release_Default();
    context = get_context(Context_id);
    old_thread = context->thr;
    Thread_Push(thread,
                old_thread);
    context->thr = thread;
    return thread; }
        
NIL_type Context_Thread_Replace(CID_type Context_id,
                                THR_type Thread)
  { Ctx_type context;
    TPL_type thread_pool;
    THR_type thread;
    context = get_context(Context_id);
    thread      = context->thr;
    thread_pool = context->tpl;
    Thread_Release(thread,
                   thread_pool);
    context->thr = Thread; }

NIL_type Context_Thread_Zap(CID_type Context_id)
  { Ctx_type context;
    THR_type thread;
    TPL_type thread_pool;
    context = get_context(Context_id);
    thread      = context->thr;
    thread_pool = context->tpl;
    Thread_Zap(&thread,
               thread_pool);
    context->thr = thread; }

/*----------------------------------------------------------------------------*/
/*---------------------- Thread procedure table management--------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------ private constants ---------------------------*/

enum { Thread_procedure_table_size = 21 };

/*------------------------------ private variables ---------------------------*/

static TPR_type Thread_procedure_table[Thread_procedure_table_size];
static UNS_type Thread_procedure_table_counter;

/*------------------------------ public functions ----------------------------*/

NIL_type Context_Proceed(CID_type Context_id)
  { Ctx_type context;
    NBR_type thread_id;
    TPR_type Procedure;
    THR_type thread;
    UNS_type raw_thread_id;
    context = get_context(Context_id);
    thread = context->thr;
    thread_id = Thread_Get_Id(thread);
    raw_thread_id = get_NBR(thread_id);
    Procedure = Thread_procedure_table[raw_thread_id];
    Procedure(Context_id); }
    
NBR_type Context_Register(TPR_type Procedure,
                          UNS_type Size)
  { NBR_type thread_id;
    if (Thread_procedure_table_counter == Thread_procedure_table_size)
      Slip_Fail(TMT_error);
    if (!Thread_Valid_Size(Size))
      Slip_Fail(TPO_error);
    Thread_procedure_table[Thread_procedure_table_counter] = Procedure;
    thread_id = make_NBR(Thread_procedure_table_counter++);
    return thread_id; }

/*------------------------------ private functions ---------------------------*/

static NIL_type thread_procedure_table_initialize(NIL_type)
  { Thread_procedure_table_counter = 0; }

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
          
NIL_type Context_Initialize_M(NIL_type)
  { Environment_Initialize_M();
    Thread_Initialize();
    context_initialize_M();
    thread_procedure_table_initialize(); }
    