                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*               memory              */
                     /*-----------------------------------*/

#include "SlipMemory.h"

/*------------------------------ private constants ---------------------------*/

static const CEL_type Busy_mask      = 0x00000002;
static const CEL_type Free_mask      = 0xFFFFFFFD;
static const CEL_type Immediate_mask = 0x00000001;
static const UNS_type Size_max       = 0x00FFFFFF;
static const CEL_type Tag_mask       = 0x000000FC;
static const CEL_type Three_bit_mask = 0x00000007;
static const CEL_type Two_bit_mask   = 0x00000003;

enum { Byte_size_shift = 2,
       Busy_shift      = 1,
       Immediate_shift = 1,
       Size_shift      = 8,
       Tag_shift       = 2 };

enum { bp_bits = 0x0,
       bP_bits = 0x1,
       Bp_bits = 0x2,
       BP_bits = 0x3 };

enum { rbp_bits = 0x0,
       rBp_bits = 0x2,
       rBP_bits = 0x3,
       Rbp_bits = 0x4,
       RBp_bits = 0x6,
       RBP_bits = 0x7 };

/*-------------------------------- opaque types ------------------------------*/

typedef union PTR { CEL_type cel;
                    PTR_type ptr; } PTR;

/*------------------------------ private variables ---------------------------*/

static PTR_type Free_pointer;
static PTR_type Head_pointer;
static PTR_type Tail_pointer;

/*------------------------------ private functions ---------------------------*/

static CEL_type busy_pointer(PTR_type Pointer)
  { return ((CEL_type)Pointer | Busy_mask); }

static CEL_type busy_cell(CEL_type Cell)
  { return (Cell | Busy_mask); }

static PTR_type free_cell(CEL_type Cell)
  { return (PTR_type)(Cell & Free_mask); }

static CEL_type free_size(UNS_type Size)
  { return (CEL_type)(Size << Size_shift); }

static SGN_type get_immediate(PTR_type Pointer)
  { return ((SGN_type)Pointer >> Immediate_shift); }

static BYT_type get_last_2_bits(CEL_type Cell)
  { return (Cell & Two_bit_mask); }

static BYT_type get_last_3_bits(CEL_type Cell)
  { return (Cell & Three_bit_mask); }

static UNS_type get_size(CEL_type Cell)
  { return (UNS_type)(Cell >> Size_shift); }

static UNS_type get_tag(CEL_type Cell)
  { return (Cell & Tag_mask) >> Tag_shift; }

static BYT_type is_busy(CEL_type Cell)
  { return (Cell & Busy_mask) >> Busy_shift; }

static BYT_type is_immediate(PTR_type Pointer)
  { return ((CEL_type)Pointer & Immediate_mask); }

static UNS_type make_byte_size(UNS_type Size)
  { return (Size + 1) << Byte_size_shift; }

static CEL_type make_header(BYT_type Tag,
                            UNS_type Size)
  { return ((CEL_type)Size << Size_shift) |
           ((CEL_type)Tag  << Tag_shift ); }

static PTR_type make_immediate(SGN_type Long)
  { CEL_type cell;
    cell = ((CEL_type)Long << Immediate_shift) | Immediate_mask;
    return (PTR_type) cell; }
    
static NIL_type sweep_and_mark(NIL_type)
  { CEL_type cell,
             header;
    PTR_type current,
             pointer;
    UNS_type size;
    for (current = Head_pointer;;)
      { cell = current->cel;                    /* <current> points to <cell> */
        switch (get_last_2_bits(cell))        /* select last 2 bits of <cell> */
          { case bp_bits:                              /* <cell> is a pointer */
              pointer = current->ptr;        /* <current> points to <pointer> */
              if (pointer < Head_pointer)        /* <pointer> is inline value */
                { current--;                             /* advance <current> */
                  continue; }
              header = pointer->cel;          /* <pointer> points to <header> */
              switch (get_last_3_bits(header))     /* last 3 bits of <header> */
                { case rbp_bits:         /* <header> is a free non-raw header */
                    pointer->cel = busy_pointer(current);    /* swap <header> */
                    current->cel = header;            /*and <pointer> content */
                    size = get_size(header);
                    if (size)                           /* if chunk not empty */
                      current = pointer + size;   /* <current> points to last */
                    else
                      current--;                         /* advance <current> */
                    continue;
                  case rBp_bits:                 /* <header> is a busy header */
                  case RBp_bits:
                    pointer = free_cell(header);
                    header = pointer->cel;
                  case Rbp_bits:             /* <header> is a free raw header */
                    pointer->cel = busy_pointer(current);
                    current->cel = header;
                    current--;                           /* advance <current> */
                    continue;
                  default: 
                    Main_Fail(IMM_error); }  
              continue;
            case Bp_bits:                          /* <cell> is a busy header */
              current = free_cell(cell);     /* restore <current> from <cell> */
              if (current == Head_pointer)  /* if backtracked to root pointer */
                return;
            case bP_bits:                       /* <cell> is an inline number */
            case BP_bits:
              current--; }}}                             /* advance <current> */

static NIL_type traverse_and_update(NIL_type)
  { CEL_type cell;
    PTR_type destination,
             free,
             pointer,
             source;
    UNS_type accumulated_size,
             size;
    source = destination = Head_pointer + 1;
    while (source < Free_pointer)
      { cell = source->cel;
        if (is_busy(cell))
          { do
              { pointer = free_cell(cell);
                cell = pointer->cel;
                pointer->ptr = destination; }
            while (is_busy(cell));
            source->cel = busy_cell(cell);
            size = get_size(cell) + 1;
            source += size;
            if (source > Free_pointer)
              Main_Fail(IMM_error);
            destination += size; }
        else
          { free = source;
            accumulated_size = 0;
            do
              { size = get_size(cell) + 1;
                if ((accumulated_size + size) >= Size_max)
                  break;
                accumulated_size += size;
                source += size;
                if (source >= Free_pointer)
                  break;
                cell = source->cel; }
            while (!is_busy(cell));
            free->cel = free_size(accumulated_size - 1); }}}

static NIL_type traverse_and_compact(NIL_type)
  { CEL_type cell;
    PTR_type destination,
             source;
    UNS_type size;
    source = destination = Head_pointer + 1;
    while (source < Free_pointer)
      { cell = (source++)->cel;
        size = get_size(cell);
        if (is_busy(cell))
          { (destination++)->ptr = free_cell(cell);
            while (size--)
              (destination++)->cel = (source++)->cel; }
        else
          source += size; }
    Free_pointer = destination; }

static NIL_type sweep_and_replace(PTR_type Original,
                                  PTR_type Replacement)
  { CEL_type cell,
             header;
    PTR_type current,
             pointer;
    UNS_type size;
    for (current = Head_pointer;;)
      { cell = current->cel;                    /* <current> points to <cell> */
        switch (get_last_2_bits(cell))        /* select last 2 bits of <cell> */
          { case bp_bits:                              /* <cell> is a pointer */
              pointer = current->ptr;        /* <current> points to <pointer> */
              if (pointer < Head_pointer)        /* <pointer> is inline value */
                { current--;                             /* advance <current> */
                  continue; }
              header = pointer->cel;          /* <pointer> points to <header> */
              switch (get_last_3_bits(header))     /* last 3 bits of <header> */
                { case rbp_bits:    /* <header> is an unmarked non-raw header */
                    size = get_size(header);
                    if (size > 0)                       /* if chunk not empty */
                      { pointer->cel = busy_pointer(current);/* swap <header> */
                        current->cel = header;       /* and <pointer> content */
                        current = pointer + size;      /* points to last cell */
                        continue; }
                  case Rbp_bits:          /* <cell> is an ummarked raw header */
                    pointer->cel = busy_cell(header);          /* mark header */
                  case rBp_bits:               /* <header> is a marked header */
                  case RBp_bits:
                    if (pointer == Original)
                      current->ptr = Replacement;
                    current--; }                         /* advance <current> */      
              continue;
            case Bp_bits:                          /* <cell> is a marked cell */
              pointer = free_cell(cell);       /* <pointer> is <cell> content */
              header = pointer->cel;          /* <pointer> points to <header> */
              current->cel = busy_cell(header);    /* restore original header */
              if (current == Original)
                pointer->ptr = Replacement;
              else
                pointer->ptr = current;           /* restore original pointer */      
              if (pointer == Head_pointer)  /* if backtracked to root pointer */
                return;
              current = pointer - 1;                     /* advance <current> */
              continue;
            case bP_bits:                       /* <cell> is an inline number */
            case BP_bits:
              current--; }}}                             /* advance <current> */

static NIL_type traverse_and_unmark(NIL_type)
  { CEL_type header;
    PTR_type source;
    UNS_type size;
    source = Head_pointer + 1;
    while (source < Free_pointer)
      { header = source->cel;
        if (is_busy(header))
          source->ptr = free_cell(header);
        size = get_size(header);
        source += size + 1; }}
                         
/*------------------------------ public functions ----------------------------*/

UNS_type Memory_Available(NIL_type)
  { return (Tail_pointer - Free_pointer); }
    
NIL_type Memory_Collect(NIL_type)
  { PTR_type root;
    root = Memory_Before_Reclaim();
    Head_pointer->ptr = root;
    sweep_and_mark();
    traverse_and_update();
    traverse_and_compact();
    root = Head_pointer->ptr;
    Memory_After_Reclaim(root); }
    
SGN_type Memory_Get_Immediate(PTR_type Pointer)
  { return get_immediate(Pointer); }

UNS_type Memory_Get_Size(PTR_type Pointer)
  { CEL_type cell;
    cell = Pointer->cel;
    return get_size(cell); }

BYT_type Memory_Get_Tag(PTR_type Pointer)
  { CEL_type cell;
    cell = Pointer->cel;
    return get_tag(cell); }

BYT_type Memory_Is_Immediate(PTR_type Pointer)
  { return is_immediate(Pointer); }

PTR_type Memory_Make_Chunk(BYT_type Tag,
                           UNS_type Size)
  { PTR_type pointer;
    PTR_type ppp;
    pointer = Free_pointer; 
    Free_pointer += Size + 1;
    pointer->cel = make_header(Tag,
                               Size);
    for (ppp = pointer + 1;
         ppp < Free_pointer;
         ppp++)
        ppp->cel = 0;
    return pointer; }

PTR_type Memory_Make_Immediate(SGN_type Long)
  { return make_immediate(Long); }
    
NIL_type Memory_Replace(PTR_type Old_pointer,
                        PTR_type New_pointer)
  { PTR_type root;
    root = Memory_Before_Reclaim();
    Head_pointer->ptr = root;
    sweep_and_replace(Old_pointer,
                      New_pointer);
    traverse_and_unmark(); }

NIL_type Memory_Set_Tag(PTR_type Pointer,
                        BYT_type Tag)
  { UNS_type size;
    size = Pointer->cel >> Size_shift;
    Pointer->cel = make_header(Tag,
                               size); }

/*----------------------------------------------------------------------------*/
                               
NIL_type Memory_Initialize(ADR_type Address,
                           UNS_type Size)
  { Head_pointer = (PTR_type)Address;
    Free_pointer = Head_pointer + 1;
    Tail_pointer = Head_pointer + Size; }